"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const rxjs_1 = require("@reactivex/rxjs");
const events_1 = require("events");
const opentracing_1 = require("opentracing");
const path = require("path");
const url = require("url");
const logging_1 = require("./logging");
/**
 * Matches:
 *
 *     /foo/node_modules/(bar)/index.d.ts
 *     /foo/node_modules/bar/node_modules/(baz)/index.d.ts
 *     /foo/node_modules/(@types/bar)/index.ts
 */
const NODE_MODULES_PACKAGE_NAME_REGEXP = /.*\/node_modules\/((?:@[^\/]+\/)?[^\/]+)\/.*$/;
/**
 * Returns the name of a package that a file is contained in
 */
function extractNodeModulesPackageName(uri) {
    const match = decodeURIComponent(url.parse(uri).pathname || '').match(NODE_MODULES_PACKAGE_NAME_REGEXP);
    return match ? match[1] : undefined;
}
exports.extractNodeModulesPackageName = extractNodeModulesPackageName;
/**
 * Matches:
 *
 *     /foo/types/(bar)/index.d.ts
 *     /foo/types/bar/node_modules/(baz)/index.d.ts
 *     /foo/types/(@types/bar)/index.ts
 */
const DEFINITELY_TYPED_PACKAGE_NAME_REGEXP = /\/types\/((?:@[^\/]+\/)?[^\/]+)\/.*$/;
/**
 * Returns the name of a package that a file in DefinitelyTyped defines.
 * E.g. `file:///foo/types/node/index.d.ts` -> `@types/node`
 */
function extractDefinitelyTypedPackageName(uri) {
    const match = decodeURIComponent(url.parse(uri).pathname || '').match(DEFINITELY_TYPED_PACKAGE_NAME_REGEXP);
    return match ? '@types/' + match[1] : undefined;
}
exports.extractDefinitelyTypedPackageName = extractDefinitelyTypedPackageName;
class PackageManager extends events_1.EventEmitter {
    constructor(updater, inMemoryFileSystem, logger = new logging_1.NoopLogger()) {
        super();
        this.updater = updater;
        this.inMemoryFileSystem = inMemoryFileSystem;
        this.logger = logger;
        /**
         * Map of package.json URIs _defined_ in the workspace to optional content.
         * Does not include package.jsons of dependencies.
         * Updated as new package.jsons are discovered.
         */
        this.packages = new Map();
        /**
         * Subscriptions to unsubscribe from on object disposal
         */
        this.subscriptions = new rxjs_1.Subscription();
        let rootPackageJsonLevel = Infinity;
        // Find locations of package.jsons _not_ inside node_modules
        this.subscriptions.add(rxjs_1.Observable.fromEvent(this.inMemoryFileSystem, 'add', Array.of)
            .subscribe(([uri, content]) => {
            const parts = url.parse(uri);
            if (!parts.pathname || !parts.pathname.endsWith('/package.json') || parts.pathname.includes('/node_modules/')) {
                return;
            }
            let parsed;
            if (content) {
                try {
                    parsed = JSON.parse(content);
                }
                catch (err) {
                    logger.error(`Error parsing package.json:`, err);
                }
            }
            // Don't override existing content with undefined
            if (parsed || !this.packages.get(uri)) {
                this.packages.set(uri, parsed);
                this.emit('parsed', uri, parsed);
            }
            this.logger.log(`Found package ${uri}`);
            // If the current root package.json is further nested than this one, replace it
            const level = parts.pathname.split('/').length;
            if (level < rootPackageJsonLevel) {
                this.rootPackageJsonUri = uri;
                rootPackageJsonLevel = level;
            }
        }));
    }
    dispose() {
        this.subscriptions.unsubscribe();
    }
    on(event, listener) {
        return super.on(event, listener);
    }
    /**
     * Returns an Iterable for all package.jsons in the workspace
     */
    packageJsonUris() {
        return this.packages.keys();
    }
    /**
     * Gets the content of the closest package.json known to to the DependencyManager in the ancestors of a URI
     */
    getClosestPackageJson(uri, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.updater.ensureStructure();
            const packageJsonUri = this.getClosestPackageJsonUri(uri);
            if (!packageJsonUri) {
                return undefined;
            }
            return yield this.getPackageJson(packageJsonUri);
        });
    }
    /**
     * Returns the parsed package.json of the passed URI
     *
     * @param uri URI of the package.json
     */
    getPackageJson(uri, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            if (uri.includes('/node_modules/')) {
                throw new Error(`Not an own package.json: ${uri}`);
            }
            let packageJson = this.packages.get(uri);
            if (packageJson) {
                return packageJson;
            }
            yield this.updater.ensure(uri, span);
            packageJson = this.packages.get(uri);
            if (!packageJson) {
                throw new Error(`Expected ${uri} to be registered in PackageManager`);
            }
            return packageJson;
        });
    }
    /**
     * Walks the parent directories of a given URI to find the first package.json that is known to the InMemoryFileSystem
     *
     * @param uri URI of a file or directory in the workspace
     * @return The found package.json or undefined if none found
     */
    getClosestPackageJsonUri(uri) {
        const parts = url.parse(uri);
        while (true) {
            if (!parts.pathname) {
                return undefined;
            }
            const packageJsonUri = url.format(Object.assign({}, parts, { pathname: path.posix.join(parts.pathname, 'package.json') }));
            if (this.packages.has(packageJsonUri)) {
                return packageJsonUri;
            }
            if (parts.pathname === '/') {
                return undefined;
            }
            parts.pathname = path.posix.dirname(parts.pathname);
        }
    }
}
exports.PackageManager = PackageManager;
//# sourceMappingURL=packages.js.map