"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const rxjs_1 = require("@reactivex/rxjs");
const iterare_1 = require("iterare");
const lodash_1 = require("lodash");
const opentracing_1 = require("opentracing");
const path = require("path");
const ts = require("typescript");
const vscode_languageserver_1 = require("vscode-languageserver");
const ast_1 = require("./ast");
const diagnostics_1 = require("./diagnostics");
const fs_1 = require("./fs");
const logging_1 = require("./logging");
const memfs_1 = require("./memfs");
const project_manager_1 = require("./project-manager");
const util_1 = require("./util");
const hashObject = require("object-hash");
const lodash_2 = require("lodash");
const url = require("url");
const packages_1 = require("./packages");
/**
 * Handles incoming requests and return responses. There is a one-to-one-to-one
 * correspondence between TCP connection, TypeScriptService instance, and
 * language workspace. TypeScriptService caches data from the compiler across
 * requests. The lifetime of the TypeScriptService instance is tied to the
 * lifetime of the TCP connection, so its caches are deleted after the
 * connection is torn down.
 *
 * Methods are camelCase versions of the LSP spec methods and dynamically
 * dispatched. Methods not to be exposed over JSON RPC are prefixed with an
 * underscore.
 */
class TypeScriptService {
    constructor(client, options = {}) {
        this.client = client;
        this.options = options;
        this.logger = new logging_1.LSPLogger(client);
    }
    /**
     * The initialize request is sent as the first request from the client to the server. If the
     * server receives request or notification before the `initialize` request it should act as
     * follows:
     *
     * - for a request the respond should be errored with `code: -32002`. The message can be picked by
     * the server.
     * - notifications should be dropped, except for the exit notification. This will allow the exit a
     * server without an initialize request.
     *
     * Until the server has responded to the `initialize` request with an `InitializeResult` the
     * client must not sent any additional requests or notifications to the server.
     *
     * During the `initialize` request the server is allowed to sent the notifications
     * `window/showMessage`, `window/logMessage` and `telemetry/event` as well as the
     * `window/showMessageRequest` request to the client.
     *
     * @return Observable of JSON Patches that build an `InitializeResult`
     */
    initialize(params, span = new opentracing_1.Span()) {
        if (params.rootUri || params.rootPath) {
            this.root = params.rootPath || util_1.uri2path(params.rootUri);
            this.rootUri = params.rootUri || util_1.path2uri('', params.rootPath);
            this._initializeFileSystems(!this.options.strict && !(params.capabilities.xcontentProvider && params.capabilities.xfilesProvider));
            this.updater = new fs_1.FileSystemUpdater(this.fileSystem, this.inMemoryFileSystem);
            this.projectManager = new project_manager_1.ProjectManager(this.root, this.inMemoryFileSystem, this.updater, !!this.options.strict, this.traceModuleResolution, this.logger);
            this.packageManager = new packages_1.PackageManager(this.updater, this.inMemoryFileSystem, this.logger);
            // Detect DefinitelyTyped
            // Fetch root package.json (if exists)
            const normRootUri = this.rootUri.endsWith('/') ? this.rootUri : this.rootUri + '/';
            const packageJsonUri = normRootUri + 'package.json';
            this.isDefinitelyTyped = rxjs_1.Observable.from(this.packageManager.getPackageJson(packageJsonUri, span))
                .map(packageJson => packageJson.name === 'definitely-typed')
                .catch(err => [false])
                .publishReplay()
                .refCount();
            // Pre-fetch files in the background if not DefinitelyTyped
            this.isDefinitelyTyped
                .mergeMap(isDefinitelyTyped => {
                if (!isDefinitelyTyped) {
                    return this.projectManager.ensureOwnFiles(span);
                }
                return [];
            })
                .subscribe(undefined, err => {
                this.logger.error(err);
            });
        }
        const result = {
            capabilities: {
                // Tell the client that the server works in FULL text document sync mode
                textDocumentSync: vscode_languageserver_1.TextDocumentSyncKind.Full,
                hoverProvider: true,
                signatureHelpProvider: {
                    triggerCharacters: ['(', ',']
                },
                definitionProvider: true,
                referencesProvider: true,
                documentSymbolProvider: true,
                workspaceSymbolProvider: true,
                xworkspaceReferencesProvider: true,
                xdefinitionProvider: true,
                xdependenciesProvider: true,
                completionProvider: {
                    resolveProvider: false,
                    triggerCharacters: ['.']
                },
                xpackagesProvider: true
            }
        };
        return rxjs_1.Observable.of({
            op: 'add',
            path: '',
            value: result
        });
    }
    /**
     * Initializes the remote file system and in-memory file system.
     * Can be overridden
     *
     * @param accessDisk Whether the language server is allowed to access the local file system
     */
    _initializeFileSystems(accessDisk) {
        this.fileSystem = accessDisk ? new fs_1.LocalFileSystem(util_1.uri2path(this.root)) : new fs_1.RemoteFileSystem(this.client);
        this.inMemoryFileSystem = new memfs_1.InMemoryFileSystem(this.root);
    }
    /**
     * The shutdown request is sent from the client to the server. It asks the server to shut down,
     * but to not exit (otherwise the response might not be delivered correctly to the client).
     * There is a separate exit notification that asks the server to exit.
     *
     * @return Observable of JSON Patches that build a `null` result
     */
    shutdown(params = {}, span = new opentracing_1.Span()) {
        this.projectManager.dispose();
        this.packageManager.dispose();
        return rxjs_1.Observable.of({ op: 'add', path: '', value: null });
    }
    /**
     * The goto definition request is sent from the client to the server to resolve the definition
     * location of a symbol at a given text document position.
     *
     * @return Observable of JSON Patches that build a `Location[]` result
     */
    textDocumentDefinition(params, span = new opentracing_1.Span()) {
        return this._getDefinitionLocations(params, span)
            .map((location) => ({ op: 'add', path: '/-', value: location }))
            .startWith({ op: 'add', path: '', value: [] });
    }
    /**
     * Returns an Observable of all definition locations found for a symbol.
     */
    _getDefinitionLocations(params, span = new opentracing_1.Span()) {
        const uri = util_1.normalizeUri(params.textDocument.uri);
        // Fetch files needed to resolve definition
        return this.projectManager.ensureReferencedFiles(uri, undefined, undefined, span)
            .toArray()
            .mergeMap(() => {
            const fileName = util_1.uri2path(uri);
            const configuration = this.projectManager.getConfiguration(fileName);
            configuration.ensureBasicFiles(span);
            const sourceFile = this._getSourceFile(configuration, fileName, span);
            if (!sourceFile) {
                throw new Error(`Expected source file ${fileName} to exist`);
            }
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, params.position.line, params.position.character);
            const definitions = configuration.getService().getDefinitionAtPosition(fileName, offset);
            return rxjs_1.Observable.from(definitions || [])
                .map((definition) => {
                const sourceFile = this._getSourceFile(configuration, definition.fileName, span);
                if (!sourceFile) {
                    throw new Error('expected source file "' + definition.fileName + '" to exist in configuration');
                }
                const start = ts.getLineAndCharacterOfPosition(sourceFile, definition.textSpan.start);
                const end = ts.getLineAndCharacterOfPosition(sourceFile, definition.textSpan.start + definition.textSpan.length);
                return {
                    uri: this._defUri(definition.fileName),
                    range: {
                        start,
                        end
                    }
                };
            });
        });
    }
    /**
     * This method is the same as textDocument/definition, except that:
     *
     * - The method returns metadata about the definition (the same metadata that
     * workspace/xreferences searches for).
     * - The concrete location to the definition (location field)
     * is optional. This is useful because the language server might not be able to resolve a goto
     * definition request to a concrete location (e.g. due to lack of dependencies) but still may
     * know some information about it.
     *
     * @return Observable of JSON Patches that build a `SymbolLocationInformation[]` result
     */
    textDocumentXdefinition(params, span = new opentracing_1.Span()) {
        return this._getSymbolLocationInformations(params, span)
            .map(symbol => ({ op: 'add', path: '/-', value: symbol }))
            .startWith({ op: 'add', path: '', value: [] });
    }
    /**
     * Returns an Observable of SymbolLocationInformations for the definition of a symbol at the given position
     */
    _getSymbolLocationInformations(params, span = new opentracing_1.Span()) {
        const uri = util_1.normalizeUri(params.textDocument.uri);
        // Ensure files needed to resolve SymbolLocationInformation are fetched
        return this.projectManager.ensureReferencedFiles(uri, undefined, undefined, span)
            .toArray()
            .mergeMap(() => {
            // Convert URI to file path
            const fileName = util_1.uri2path(uri);
            // Get closest tsconfig configuration
            const configuration = this.projectManager.getConfiguration(fileName);
            configuration.ensureBasicFiles(span);
            const sourceFile = this._getSourceFile(configuration, fileName, span);
            if (!sourceFile) {
                throw new Error(`Unknown text document ${uri}`);
            }
            // Convert line/character to offset
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, params.position.line, params.position.character);
            // Query TypeScript for references
            return rxjs_1.Observable.from(configuration.getService().getDefinitionAtPosition(fileName, offset) || [])
                .mergeMap((definition) => {
                const definitionUri = this._defUri(definition.fileName);
                // Get the PackageDescriptor
                return this._getPackageDescriptor(definitionUri)
                    .map((packageDescriptor) => {
                    const sourceFile = this._getSourceFile(configuration, definition.fileName, span);
                    if (!sourceFile) {
                        throw new Error(`Expected source file ${definition.fileName} to exist in configuration`);
                    }
                    const symbol = util_1.defInfoToSymbolDescriptor(definition);
                    if (packageDescriptor) {
                        symbol.package = packageDescriptor;
                    }
                    return {
                        symbol,
                        location: {
                            uri: definitionUri,
                            range: {
                                start: ts.getLineAndCharacterOfPosition(sourceFile, definition.textSpan.start),
                                end: ts.getLineAndCharacterOfPosition(sourceFile, definition.textSpan.start + definition.textSpan.length)
                            }
                        }
                    };
                });
            });
        });
    }
    /**
     * Finds the PackageDescriptor a given file belongs to
     *
     * @return Observable that emits a single PackageDescriptor or undefined if the definition does not belong to any package
     */
    _getPackageDescriptor(uri) {
        // Get package name of the dependency in which the symbol is defined in, if any
        const packageName = packages_1.extractNodeModulesPackageName(uri);
        if (packageName) {
            // The symbol is part of a dependency in node_modules
            // Build URI to package.json of the Dependency
            const encodedPackageName = packageName.split('/').map(encodeURIComponent).join('/');
            const parts = url.parse(uri);
            const packageJsonUri = url.format(Object.assign({}, parts, { pathname: parts.pathname.slice(0, parts.pathname.lastIndexOf('/node_modules/' + encodedPackageName)) + `/node_modules/${encodedPackageName}/package.json` }));
            // Fetch the package.json of the dependency
            return rxjs_1.Observable.from(this.updater.ensure(packageJsonUri))
                .map(() => {
                const packageJson = JSON.parse(this.inMemoryFileSystem.getContent(packageJsonUri));
                const { name, version } = packageJson;
                if (name) {
                    // Used by the LSP proxy to shortcut database lookup of repo URL for PackageDescriptor
                    let repoURL;
                    if (name.startsWith('@types/')) {
                        // if the dependency package is an @types/ package, point the repo to DefinitelyTyped
                        repoURL = 'https://github.com/DefinitelyTyped/DefinitelyTyped';
                    }
                    else {
                        // else use repository field from package.json
                        repoURL = typeof packageJson.repository === 'object' ? packageJson.repository.url : undefined;
                    }
                    return { name, version, repoURL };
                }
                return undefined;
            });
        }
        else {
            // The symbol is defined in the root package of the workspace, not in a dependency
            // Get root package.json
            return rxjs_1.Observable.from(this.packageManager.getClosestPackageJson(uri))
                .map((packageJson) => {
                if (!packageJson) {
                    // Workspace has no package.json
                    return undefined;
                }
                let { name, version } = packageJson;
                if (name) {
                    let repoURL = typeof packageJson.repository === 'object' ? packageJson.repository.url : undefined;
                    // If the root package is DefinitelyTyped, find out the proper @types package name for each typing
                    if (name === 'definitely-typed') {
                        name = packages_1.extractDefinitelyTypedPackageName(uri);
                        if (!name) {
                            this.logger.error(`Could not extract package name from DefinitelyTyped URI ${uri}`);
                            return undefined;
                        }
                        version = undefined;
                        repoURL = 'https://github.com/DefinitelyTyped/DefinitelyTyped';
                    }
                    return { name, version, repoURL };
                }
                return undefined;
            });
        }
    }
    /**
     * The hover request is sent from the client to the server to request hover information at a
     * given text document position.
     *
     * @return Observable of JSON Patches that build a `Hover` result
     */
    textDocumentHover(params, span = new opentracing_1.Span()) {
        return this._getHover(params, span)
            .map(hover => ({ op: 'add', path: '', value: hover }));
    }
    /**
     * Returns an Observable for a Hover at the given position
     */
    _getHover(params, span = new opentracing_1.Span()) {
        const uri = util_1.normalizeUri(params.textDocument.uri);
        // Ensure files needed to resolve hover are fetched
        return this.projectManager.ensureReferencedFiles(uri, undefined, undefined, span)
            .toArray()
            .map(() => {
            const fileName = util_1.uri2path(uri);
            const configuration = this.projectManager.getConfiguration(fileName);
            configuration.ensureBasicFiles(span);
            const sourceFile = this._getSourceFile(configuration, fileName, span);
            if (!sourceFile) {
                throw new Error(`Unknown text document ${uri}`);
            }
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, params.position.line, params.position.character);
            const info = configuration.getService().getQuickInfoAtPosition(fileName, offset);
            if (!info) {
                return { contents: [] };
            }
            const contents = [];
            // Add declaration without the kind
            const declaration = ts.displayPartsToString(info.displayParts).replace(/^\(.+\)\s+/, '');
            contents.push({ language: 'typescript', value: declaration });
            // Add kind with modifiers, e.g. "method (private, ststic)", "class (exported)"
            if (info.kind) {
                let kind = '**' + info.kind + '**';
                const modifiers = info.kindModifiers
                    .split(',')
                    .filter(mod => mod && (mod !== ts.ScriptElementKindModifier.exportedModifier
                    || info.kind !== ts.ScriptElementKind.constructorImplementationElement))
                    .map(mod => ({
                    [ts.ScriptElementKindModifier.ambientModifier]: 'ambient',
                    [ts.ScriptElementKindModifier.exportedModifier]: 'exported'
                })[mod] || mod);
                if (modifiers.length > 0) {
                    kind += ' _(' + modifiers.join(', ') + ')_';
                }
                contents.push(kind);
            }
            // Add documentation
            const documentation = ts.displayPartsToString(info.documentation);
            if (documentation) {
                contents.push(documentation);
            }
            const start = ts.getLineAndCharacterOfPosition(sourceFile, info.textSpan.start);
            const end = ts.getLineAndCharacterOfPosition(sourceFile, info.textSpan.start + info.textSpan.length);
            return {
                contents,
                range: {
                    start,
                    end
                }
            };
        });
    }
    /**
     * The references request is sent from the client to the server to resolve project-wide
     * references for the symbol denoted by the given text document position.
     *
     * Returns all references to the symbol at the position in the own workspace, including references inside node_modules.
     *
     * @return Observable of JSON Patches that build a `Location[]` result
     */
    textDocumentReferences(params, span = new opentracing_1.Span()) {
        const uri = util_1.normalizeUri(params.textDocument.uri);
        // Ensure all files were fetched to collect all references
        return rxjs_1.Observable.from(this.projectManager.ensureOwnFiles(span))
            .mergeMap(() => {
            // Convert URI to file path because TypeScript doesn't work with URIs
            const fileName = util_1.uri2path(uri);
            // Get tsconfig configuration for requested file
            const configuration = this.projectManager.getConfiguration(fileName);
            // Ensure all files have been added
            configuration.ensureAllFiles(span);
            const program = configuration.getProgram(span);
            if (!program) {
                return [];
            }
            // Get SourceFile object for requested file
            const sourceFile = this._getSourceFile(configuration, fileName, span);
            if (!sourceFile) {
                throw new Error(`Source file ${fileName} does not exist`);
            }
            // Convert line/character to offset
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, params.position.line, params.position.character);
            // Request references at position from TypeScript
            // Despite the signature, getReferencesAtPosition() can return undefined
            return rxjs_1.Observable.from(configuration.getService().getReferencesAtPosition(fileName, offset) || [])
                .filter(reference => 
            // Filter declaration if not requested
            (!reference.isDefinition || (params.context && params.context.includeDeclaration))
                && !reference.fileName.includes('/node_modules/'))
                .map((reference) => {
                const sourceFile = program.getSourceFile(reference.fileName);
                if (!sourceFile) {
                    throw new Error(`Source file ${reference.fileName} does not exist`);
                }
                // Convert offset to line/character position
                const start = ts.getLineAndCharacterOfPosition(sourceFile, reference.textSpan.start);
                const end = ts.getLineAndCharacterOfPosition(sourceFile, reference.textSpan.start + reference.textSpan.length);
                return {
                    uri: util_1.path2uri(this.root, reference.fileName),
                    range: {
                        start,
                        end
                    }
                };
            });
        })
            .map((location) => ({ op: 'add', path: '/-', value: location }))
            .startWith({ op: 'add', path: '', value: [] });
    }
    /**
     * The workspace symbol request is sent from the client to the server to list project-wide
     * symbols matching the query string. The text document parameter specifies the active document
     * at time of the query. This can be used to rank or limit results.
     *
     * @return Observable of JSON Patches that build a `SymbolInformation[]` result
     */
    workspaceSymbol(params, span = new opentracing_1.Span()) {
        // Always return max. 50 results
        // TODO stream 50 results, then re-query and stream the rest
        const limit = Math.min(params.limit || Infinity, 50);
        const query = params.query;
        const symbolQuery = params.symbol && Object.assign({}, params.symbol);
        if (symbolQuery && symbolQuery.package) {
            // Strip all fields except name from PackageDescriptor
            symbolQuery.package = { name: symbolQuery.package.name };
        }
        // Return cached result for empty query, if available
        if (!query && !symbolQuery && this.emptyQueryWorkspaceSymbols) {
            return this.emptyQueryWorkspaceSymbols;
        }
        // Use special logic for DefinitelyTyped
        return this.isDefinitelyTyped
            .mergeMap(isDefinitelyTyped => {
            if (isDefinitelyTyped) {
                return this._workspaceSymbolDefinitelyTyped(Object.assign({}, params, { limit }), span);
            }
            // A workspace/symol request searches all symbols in own code, but not in dependencies
            let patches = rxjs_1.Observable.from(this.projectManager.ensureOwnFiles(span))
                .mergeMap(() => symbolQuery && symbolQuery.package
                ? rxjs_1.Observable.from(this.packageManager.packageJsonUris())
                    .filter(packageJsonUri => !symbolQuery || !symbolQuery.package || !symbolQuery.package.name || JSON.parse(this.inMemoryFileSystem.getContent(packageJsonUri)).name === symbolQuery.package.name)
                    .mergeMap(packageJsonUri => rxjs_1.Observable.merge(lodash_2.castArray(this.projectManager.getParentConfiguration(packageJsonUri) || []), 
                // Search child directories starting at the directory of the package.json
                this.projectManager.getChildConfigurations(url.resolve(packageJsonUri, '.'))))
                : this.projectManager.configurations())
                .mergeMap(config => this._collectWorkspaceSymbols(config, query || symbolQuery, limit, span))
                .filter(symbol => !symbol.location.uri.includes('/node_modules/'))
                .distinct(symbol => hashObject(symbol, { respectType: false }))
                .take(limit)
                .map(symbol => ({ op: 'add', path: '/-', value: symbol }))
                .startWith({ op: 'add', path: '', value: [] });
            if (!query && !symbolQuery) {
                patches = this.emptyQueryWorkspaceSymbols = patches.publishReplay().refCount();
            }
            return patches;
        });
    }
    /**
     * Specialised version of workspaceSymbol for DefinitelyTyped.
     * Searches only in the correct subdirectory for the given PackageDescriptor.
     * Will error if not passed a SymbolDescriptor query with an `@types` PackageDescriptor
     *
     * @return Observable of JSON Patches that build a `SymbolInformation[]` result
     */
    _workspaceSymbolDefinitelyTyped(params, childOf = new opentracing_1.Span()) {
        const span = childOf.tracer().startSpan('Handle workspace/symbol DefinitelyTyped', { childOf });
        if (!params.symbol || !params.symbol.package || !params.symbol.package.name || !params.symbol.package.name.startsWith('@types/')) {
            return rxjs_1.Observable.throw('workspace/symbol on DefinitelyTyped is only supported with a SymbolDescriptor query with an @types PackageDescriptor');
        }
        const symbolQuery = Object.assign({}, params.symbol);
        // Don't match PackageDescriptor on symbols
        symbolQuery.package = undefined;
        // Fetch all files in the package subdirectory
        // All packages are in the types/ subdirectory
        const normRootUri = this.rootUri.endsWith('/') ? this.rootUri : this.rootUri + '/';
        const packageRootUri = normRootUri + params.symbol.package.name.substr(1) + '/';
        return rxjs_1.Observable.from(this.updater.ensureStructure(span))
            .mergeMap(() => rxjs_1.Observable.from(this.inMemoryFileSystem.uris()))
            .filter(uri => uri.startsWith(packageRootUri))
            .mergeMap(uri => this.updater.ensure(uri, span))
            .toArray()
            .mergeMap(() => {
            span.log({ event: 'fetched package files' });
            // Search symbol in configuration
            // forcing TypeScript mode
            const config = this.projectManager.getConfiguration(util_1.uri2path(packageRootUri), 'ts');
            return this._collectWorkspaceSymbols(config, params.query || symbolQuery, params.limit, span);
        })
            .map(symbol => ({ op: 'add', path: '/-', value: symbol }))
            .startWith({ op: 'add', path: '', value: [] })
            .catch(err => {
            span.setTag('error', true);
            span.log({ 'event': 'error', 'error.object': err, 'message': err.message, 'stack': err.stack });
            throw err;
        })
            .finally(() => {
            span.finish();
        });
    }
    /**
     * The document symbol request is sent from the client to the server to list all symbols found
     * in a given text document.
     *
     * @return Observable of JSON Patches that build a `SymbolInformation[]` result
     */
    textDocumentDocumentSymbol(params, span = new opentracing_1.Span()) {
        const uri = util_1.normalizeUri(params.textDocument.uri);
        // Ensure files needed to resolve symbols are fetched
        return this.projectManager.ensureReferencedFiles(uri, undefined, undefined, span)
            .toArray()
            .mergeMap(() => {
            const fileName = util_1.uri2path(uri);
            const config = this.projectManager.getConfiguration(fileName);
            config.ensureBasicFiles(span);
            const sourceFile = this._getSourceFile(config, fileName, span);
            if (!sourceFile) {
                return [];
            }
            const tree = config.getService().getNavigationTree(fileName);
            return this._flattenNavigationTreeItem(tree, null, sourceFile);
        })
            .map(symbol => ({ op: 'add', path: '', value: symbol }))
            .startWith({ op: 'add', path: '', value: [] });
    }
    /**
     * The workspace references request is sent from the client to the server to locate project-wide
     * references to a symbol given its description / metadata.
     *
     * @return Observable of JSON Patches that build a `ReferenceInformation[]` result
     */
    workspaceXreferences(params, span = new opentracing_1.Span()) {
        const queryWithoutPackage = lodash_2.omit(params.query, 'package');
        return rxjs_1.Observable.from(this.projectManager.ensureAllFiles(span))
            .mergeMap(() => {
            // if we were hinted that we should only search a specific package, find it and only search the owning tsconfig.json
            if (params.hints && params.hints.dependeePackageName) {
                return rxjs_1.Observable.from(this.packageManager.packageJsonUris())
                    .filter(uri => JSON.parse(this.inMemoryFileSystem.getContent(uri)).name === params.hints.dependeePackageName)
                    .take(1)
                    .mergeMap(uri => {
                    const config = this.projectManager.getParentConfiguration(uri);
                    if (!config) {
                        return this.projectManager.configurations();
                    }
                    return [config];
                });
            }
            // else search all tsconfig.jsons
            return this.projectManager.configurations();
        })
            .mergeMap((config) => {
            config.ensureAllFiles(span);
            const program = config.getProgram(span);
            if (!program) {
                return rxjs_1.Observable.empty();
            }
            return rxjs_1.Observable.from(program.getSourceFiles())
                .filter(source => !util_1.toUnixPath(source.fileName).includes('/node_modules/'))
                .mergeMap(source => 
            // Iterate AST of source file
            rxjs_1.Observable.from(ast_1.walkMostAST(source))
                .filter((node) => node.kind === ts.SyntaxKind.Identifier)
                .mergeMap(node => {
                try {
                    // Get DefinitionInformations at the node
                    return rxjs_1.Observable.from(config.getService().getDefinitionAtPosition(source.fileName, node.pos + 1) || [])
                        .mergeMap(definition => {
                        const symbol = util_1.defInfoToSymbolDescriptor(definition);
                        // Check if SymbolDescriptor without PackageDescriptor matches
                        if (!util_1.isSymbolDescriptorMatch(queryWithoutPackage, symbol)) {
                            return [];
                        }
                        // If no PackageDescriptor query, return match
                        if (!params.query.package || !params.query.package) {
                            return [symbol];
                        }
                        // If SymbolDescriptor matched and the query contains a PackageDescriptor, get package.json and match PackageDescriptor name
                        // TODO match full PackageDescriptor (version)
                        const uri = util_1.path2uri('', definition.fileName);
                        return this._getPackageDescriptor(uri)
                            .mergeMap(packageDescriptor => {
                            symbol.package = packageDescriptor;
                            return packageDescriptor && packageDescriptor.name === params.query.package.name
                                ? [symbol]
                                : [];
                        });
                    })
                        .map(symbol => ({
                        symbol,
                        reference: {
                            uri: this._defUri(source.fileName),
                            range: {
                                start: ts.getLineAndCharacterOfPosition(source, node.pos + 1),
                                end: ts.getLineAndCharacterOfPosition(source, node.end)
                            }
                        }
                    }));
                }
                catch (err) {
                    // Continue with next node on error
                    // Workaround for https://github.com/Microsoft/TypeScript/issues/15219
                    this.logger.error(`workspace/xreferences: Error getting definition for ${source.fileName} at offset ${node.pos + 1}`, err);
                    span.log({ 'event': 'error', 'error.object': err, 'message': err.message, 'stack': err.stack });
                    return [];
                }
            }));
        })
            .map((reference) => ({ op: 'add', path: '/-', value: reference }))
            .startWith({ op: 'add', path: '', value: [] });
    }
    /**
     * This method returns metadata about the package(s) defined in a workspace and a list of
     * dependencies for each package.
     *
     * This method is necessary to implement cross-repository jump-to-def when it is not possible to
     * resolve the global location of the definition from data present or derived from the local
     * workspace. For example, a package manager might not include information about the source
     * repository of each dependency. In this case, definition resolution requires mapping from
     * package descriptor to repository revision URL. A reverse index can be constructed from calls
     * to workspace/xpackages to provide an efficient mapping.
     *
     * @return Observable of JSON Patches that build a `PackageInformation[]` result
     */
    workspaceXpackages(params = {}, span = new opentracing_1.Span()) {
        // Ensure package.json files
        return rxjs_1.Observable.from(this.projectManager.ensureModuleStructure(span))
            .mergeMap(() => this.inMemoryFileSystem.uris())
            .filter(uri => uri.includes('/package.json') && !uri.includes('/node_modules/'))
            .mergeMap(uri => this.packageManager.getPackageJson(uri))
            .mergeMap(packageJson => {
            if (!packageJson.name) {
                return [];
            }
            const packageDescriptor = {
                name: packageJson.name,
                version: packageJson.version,
                repoURL: typeof packageJson.repository === 'object' && packageJson.repository.url || undefined
            };
            // Collect all dependencies for this package.json
            return rxjs_1.Observable.of('dependencies', 'devDependencies', 'peerDependencies', 'optionalDependencies')
                .filter(key => !!packageJson[key])
                .mergeMap(key => lodash_1.toPairs(packageJson[key]))
                .map(([name, version]) => ({
                attributes: {
                    name,
                    version
                },
                hints: {
                    dependeePackageName: packageJson.name
                }
            }))
                .toArray()
                .map((dependencies) => ({
                package: packageDescriptor,
                dependencies
            }));
        })
            .map((packageInfo) => ({ op: 'add', path: '/-', value: packageInfo }))
            .startWith({ op: 'add', path: '', value: [] });
    }
    /**
     * Returns all dependencies of a workspace.
     * Superseded by workspace/xpackages
     *
     * @return Observable of JSON Patches that build a `DependencyReference[]` result
     */
    workspaceXdependencies(params = {}, span = new opentracing_1.Span()) {
        // Ensure package.json files
        return rxjs_1.Observable.from(this.projectManager.ensureModuleStructure())
            .mergeMap(() => this.inMemoryFileSystem.uris())
            .filter(uri => uri.includes('/package.json') && !uri.includes('/node_modules/'))
            .mergeMap(uri => this.packageManager.getPackageJson(uri))
            .mergeMap(packageJson => rxjs_1.Observable.of('dependencies', 'devDependencies', 'peerDependencies', 'optionalDependencies')
            .filter(key => !!packageJson[key])
            .mergeMap(key => lodash_1.toPairs(packageJson[key]))
            .map(([name, version]) => ({
            attributes: {
                name,
                version
            },
            hints: {
                dependeePackageName: packageJson.name
            }
        })))
            .map((dependency) => ({ op: 'add', path: '/-', value: dependency }))
            .startWith({ op: 'add', path: '', value: [] });
    }
    /**
     * The Completion request is sent from the client to the server to compute completion items at a
     * given cursor position. Completion items are presented in the
     * [IntelliSense](https://code.visualstudio.com/docs/editor/editingevolved#_intellisense) user
     * interface. If computing full completion items is expensive, servers can additionally provide
     * a handler for the completion item resolve request ('completionItem/resolve'). This request is
     * sent when a completion item is selected in the user interface. A typically use case is for
     * example: the 'textDocument/completion' request doesn't fill in the `documentation` property
     * for returned completion items since it is expensive to compute. When the item is selected in
     * the user interface then a 'completionItem/resolve' request is sent with the selected
     * completion item as a param. The returned completion item should have the documentation
     * property filled in.
     *
     * @return Observable of JSON Patches that build a `CompletionList` result
     */
    textDocumentCompletion(params, span = new opentracing_1.Span()) {
        const uri = util_1.normalizeUri(params.textDocument.uri);
        // Ensure files needed to suggest completions are fetched
        return this.projectManager.ensureReferencedFiles(uri, undefined, undefined, span)
            .toArray()
            .mergeMap(() => {
            const fileName = util_1.uri2path(uri);
            const configuration = this.projectManager.getConfiguration(fileName);
            configuration.ensureBasicFiles(span);
            const sourceFile = this._getSourceFile(configuration, fileName, span);
            if (!sourceFile) {
                return [];
            }
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, params.position.line, params.position.character);
            const completions = configuration.getService().getCompletionsAtPosition(fileName, offset);
            if (completions == null) {
                return [];
            }
            return rxjs_1.Observable.from(completions.entries)
                .map(entry => {
                const item = { label: entry.name };
                const kind = completionKinds[entry.kind];
                if (kind) {
                    item.kind = kind;
                }
                if (entry.sortText) {
                    item.sortText = entry.sortText;
                }
                const details = configuration.getService().getCompletionEntryDetails(fileName, offset, entry.name);
                if (details) {
                    item.documentation = ts.displayPartsToString(details.documentation);
                    item.detail = ts.displayPartsToString(details.displayParts);
                }
                return { op: 'add', path: '/items/-', value: item };
            })
                .startWith({ op: 'add', path: '/isIncomplete', value: false });
        })
            .startWith({ op: 'add', path: '', value: { isIncomplete: true, items: [] } });
    }
    /**
     * The signature help request is sent from the client to the server to request signature
     * information at a given cursor position.
     *
     * @return Observable of JSON Patches that build a `SignatureHelp` result
     */
    textDocumentSignatureHelp(params, span = new opentracing_1.Span()) {
        const uri = util_1.normalizeUri(params.textDocument.uri);
        // Ensure files needed to resolve signature are fetched
        return this.projectManager.ensureReferencedFiles(uri, undefined, undefined, span)
            .toArray()
            .map(() => {
            const filePath = util_1.uri2path(uri);
            const configuration = this.projectManager.getConfiguration(filePath);
            configuration.ensureBasicFiles(span);
            const sourceFile = this._getSourceFile(configuration, filePath, span);
            if (!sourceFile) {
                throw new Error(`expected source file ${filePath} to exist in configuration`);
            }
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, params.position.line, params.position.character);
            const signatures = configuration.getService().getSignatureHelpItems(filePath, offset);
            if (!signatures) {
                return { signatures: [], activeParameter: 0, activeSignature: 0 };
            }
            const signatureInformations = signatures.items.map((item) => {
                const prefix = ts.displayPartsToString(item.prefixDisplayParts);
                const params = item.parameters.map(p => ts.displayPartsToString(p.displayParts)).join(', ');
                const suffix = ts.displayPartsToString(item.suffixDisplayParts);
                const parameters = item.parameters.map((p) => ({
                    label: ts.displayPartsToString(p.displayParts),
                    documentation: ts.displayPartsToString(p.documentation)
                }));
                return {
                    label: prefix + params + suffix,
                    documentation: ts.displayPartsToString(item.documentation),
                    parameters
                };
            });
            return {
                signatures: signatureInformations,
                activeSignature: signatures.selectedItemIndex,
                activeParameter: signatures.argumentIndex
            };
        })
            .map(signatureHelp => ({ op: 'add', path: '', value: signatureHelp }));
    }
    /**
     * The document open notification is sent from the client to the server to signal newly opened
     * text documents. The document's truth is now managed by the client and the server must not try
     * to read the document's truth using the document's uri.
     */
    textDocumentDidOpen(params) {
        return __awaiter(this, void 0, void 0, function* () {
            const uri = util_1.normalizeUri(params.textDocument.uri);
            // Ensure files needed for most operations are fetched
            yield this.projectManager.ensureReferencedFiles(uri).toPromise();
            this.projectManager.didOpen(uri, params.textDocument.text);
            this._publishDiagnostics(uri);
        });
    }
    /**
     * The document change notification is sent from the client to the server to signal changes to a
     * text document. In 2.0 the shape of the params has changed to include proper version numbers
     * and language ids.
     */
    textDocumentDidChange(params) {
        return __awaiter(this, void 0, void 0, function* () {
            const uri = util_1.normalizeUri(params.textDocument.uri);
            let text;
            for (const change of params.contentChanges) {
                if (change.range || change.rangeLength) {
                    throw new Error('incremental updates in textDocument/didChange not supported for file ' + uri);
                }
                text = change.text;
            }
            if (!text) {
                return;
            }
            this.projectManager.didChange(uri, text);
            this._publishDiagnostics(uri);
        });
    }
    /**
     * Generates and publishes diagnostics for a given file
     *
     * @param uri URI of the file to check
     */
    _publishDiagnostics(uri, span = new opentracing_1.Span()) {
        const config = this.projectManager.getParentConfiguration(uri);
        if (!config) {
            return;
        }
        const program = config.getProgram(span);
        if (!program) {
            return;
        }
        const sourceFile = program.getSourceFile(util_1.uri2path(uri));
        if (!sourceFile) {
            return;
        }
        const tsDiagnostics = ts.getPreEmitDiagnostics(program, sourceFile);
        const diagnostics = iterare_1.default(tsDiagnostics)
            .filter(diagnostic => !!diagnostic.file)
            .map(diagnostics_1.convertTsDiagnostic)
            .toArray();
        this.client.textDocumentPublishDiagnostics({ uri, diagnostics });
    }
    /**
     * The document save notification is sent from the client to the server when the document was
     * saved in the client.
     */
    textDocumentDidSave(params) {
        return __awaiter(this, void 0, void 0, function* () {
            const uri = util_1.normalizeUri(params.textDocument.uri);
            // Ensure files needed to suggest completions are fetched
            yield this.projectManager.ensureReferencedFiles(uri).toPromise();
            this.projectManager.didSave(uri);
        });
    }
    /**
     * The document close notification is sent from the client to the server when the document got
     * closed in the client. The document's truth now exists where the document's uri points to
     * (e.g. if the document's uri is a file uri the truth now exists on disk).
     */
    textDocumentDidClose(params) {
        return __awaiter(this, void 0, void 0, function* () {
            const uri = util_1.normalizeUri(params.textDocument.uri);
            // Ensure files needed to suggest completions are fetched
            yield this.projectManager.ensureReferencedFiles(uri).toPromise();
            this.projectManager.didClose(uri);
            // Clear diagnostics
            this.client.textDocumentPublishDiagnostics({ uri, diagnostics: [] });
        });
    }
    /**
     * Fetches (or creates if needed) source file object for a given file name
     *
     * @param configuration project configuration
     * @param fileName file name to fetch source file for or create it
     * @param span Span for tracing
     */
    _getSourceFile(configuration, fileName, span = new opentracing_1.Span()) {
        let program = configuration.getProgram(span);
        if (!program) {
            return undefined;
        }
        const sourceFile = program.getSourceFile(fileName);
        if (sourceFile) {
            return sourceFile;
        }
        if (!this.projectManager.hasFile(fileName)) {
            return undefined;
        }
        configuration.getHost().addFile(fileName);
        program = configuration.getProgram(span);
        return program && program.getSourceFile(fileName);
    }
    /**
     * Returns an Iterator for all symbols in a given config that match a given SymbolDescriptor
     *
     * @param config The ProjectConfiguration to search
     * @param query A text or SymbolDescriptor query
     * @param limit An optional limit that is passed to TypeScript
     * @return Observable of SymbolInformations
     */
    _collectWorkspaceSymbols(config, query, limit = Infinity, childOf = new opentracing_1.Span()) {
        const span = childOf.tracer().startSpan('Collect workspace symbols', { childOf });
        span.addTags({ config: config.configFilePath, query, limit });
        return (() => {
            try {
                config.ensureAllFiles(span);
                const program = config.getProgram(span);
                if (!program) {
                    return rxjs_1.Observable.empty();
                }
                if (query) {
                    let items;
                    if (typeof query === 'string') {
                        // Query by text query
                        items = rxjs_1.Observable.from(config.getService().getNavigateToItems(query, limit, undefined, false));
                    }
                    else {
                        const queryWithoutPackage = lodash_2.omit(query, 'package');
                        // Query by name
                        items = rxjs_1.Observable.from(config.getService().getNavigateToItems(query.name || '', limit, undefined, false))
                            .filter(item => util_1.isSymbolDescriptorMatch(queryWithoutPackage, {
                            kind: item.kind,
                            name: item.name,
                            containerKind: item.containerKind,
                            containerName: item.containerName
                        }))
                            .mergeMap(item => {
                            if (!query.package || !query.package.name) {
                                return [item];
                            }
                            const uri = util_1.path2uri('', item.fileName);
                            return rxjs_1.Observable.from(this.packageManager.getClosestPackageJson(uri, span))
                                .mergeMap(packageJson => packageJson && packageJson.name === query.package.name ? [item] : []);
                        });
                    }
                    return rxjs_1.Observable.from(items)
                        .map(item => {
                        const sourceFile = program.getSourceFile(item.fileName);
                        if (!sourceFile) {
                            throw new Error(`Source file ${item.fileName} does not exist`);
                        }
                        const symbolInformation = {
                            name: item.name,
                            kind: util_1.convertStringtoSymbolKind(item.kind),
                            location: {
                                uri: this._defUri(item.fileName),
                                range: {
                                    start: ts.getLineAndCharacterOfPosition(sourceFile, item.textSpan.start),
                                    end: ts.getLineAndCharacterOfPosition(sourceFile, item.textSpan.start + item.textSpan.length)
                                }
                            }
                        };
                        if (item.containerName) {
                            symbolInformation.containerName = item.containerName;
                        }
                        return symbolInformation;
                    })
                        .filter(symbolInformation => util_1.isLocalUri(symbolInformation.location.uri));
                }
                else {
                    // An empty query uses a different algorithm to iterate all files and aggregate the symbols per-file to get all symbols
                    // TODO make all implementations use this? It has the advantage of being streamable and cancellable
                    return rxjs_1.Observable.from(this._getNavigationTreeItems(config)).take(limit);
                }
            }
            catch (err) {
                return rxjs_1.Observable.throw(err);
            }
        })()
            .do(lodash_2.noop, err => {
            span.setTag('error', true);
            span.log({ 'event': 'error', 'error.object': err, 'stack': err.stack, 'message': err.message });
        })
            .finally(() => {
            span.finish();
        });
    }
    /**
     * Transforms definition's file name to URI. If definition belongs to TypeScript library,
     * returns git://github.com/Microsoft/TypeScript URL, otherwise returns file:// one
     */
    _defUri(filePath) {
        if (memfs_1.isTypeScriptLibrary(filePath)) {
            return 'git://github.com/Microsoft/TypeScript?v' + ts.version + '#lib/' + path.basename(filePath);
        }
        return util_1.path2uri(this.root, filePath);
    }
    /**
     * Fetches up to limit navigation bar items from given project, flattens them
     */
    _getNavigationTreeItems(configuration, span = new opentracing_1.Span()) {
        const program = configuration.getProgram(span);
        if (!program) {
            return iterare_1.default([]);
        }
        return iterare_1.default(program.getSourceFiles())
            .filter(sourceFile => !memfs_1.isTypeScriptLibrary(sourceFile.fileName))
            .map(sourceFile => {
            try {
                const tree = configuration.getService().getNavigationTree(sourceFile.fileName);
                return this._flattenNavigationTreeItem(tree, null, sourceFile);
            }
            catch (e) {
                this.logger.error('Could not get navigation tree for file', sourceFile.fileName);
                return [];
            }
        })
            .flatten();
    }
    /**
     * Flattens navigation tree by emitting acceptable NavigationTreeItems as SymbolInformations.
     * Some items (source files, modules) may be excluded
     */
    *_flattenNavigationTreeItem(item, parent, sourceFile) {
        const acceptable = TypeScriptService.isAcceptableNavigationTreeItem(item);
        if (acceptable) {
            const span = item.spans[0];
            const symbolInformation = {
                name: item.text,
                kind: util_1.convertStringtoSymbolKind(item.kind),
                location: {
                    uri: this._defUri(sourceFile.fileName),
                    range: {
                        start: ts.getLineAndCharacterOfPosition(sourceFile, span.start),
                        end: ts.getLineAndCharacterOfPosition(sourceFile, span.start + span.length)
                    }
                }
            };
            if (parent) {
                symbolInformation.containerName = parent.text;
            }
            yield symbolInformation;
        }
        if (item.childItems) {
            for (const childItem of item.childItems) {
                yield* this._flattenNavigationTreeItem(childItem, acceptable ? item : null, sourceFile);
            }
        }
    }
    /**
     * @return true if navigation tree item is acceptable for inclusion into workspace/symbols
     */
    static isAcceptableNavigationTreeItem(item) {
        // modules and source files should be excluded
        if ([ts.ScriptElementKind.moduleElement, 'sourcefile'].indexOf(item.kind) >= 0) {
            return false;
        }
        // special items may start with ", (, [, or <
        if (/^[<\(\[\"]/.test(item.text)) {
            return false;
        }
        // magic words
        if (['default', 'constructor', 'new()'].indexOf(item.text) >= 0) {
            return false;
        }
        return true;
    }
}
exports.TypeScriptService = TypeScriptService;
/**
 * Maps string-based CompletionEntry::kind to enum-based CompletionItemKind
 */
const completionKinds = {
    class: vscode_languageserver_1.CompletionItemKind.Class,
    constructor: vscode_languageserver_1.CompletionItemKind.Constructor,
    enum: vscode_languageserver_1.CompletionItemKind.Enum,
    field: vscode_languageserver_1.CompletionItemKind.Field,
    file: vscode_languageserver_1.CompletionItemKind.File,
    function: vscode_languageserver_1.CompletionItemKind.Function,
    interface: vscode_languageserver_1.CompletionItemKind.Interface,
    keyword: vscode_languageserver_1.CompletionItemKind.Keyword,
    method: vscode_languageserver_1.CompletionItemKind.Method,
    module: vscode_languageserver_1.CompletionItemKind.Module,
    property: vscode_languageserver_1.CompletionItemKind.Property,
    reference: vscode_languageserver_1.CompletionItemKind.Reference,
    snippet: vscode_languageserver_1.CompletionItemKind.Snippet,
    text: vscode_languageserver_1.CompletionItemKind.Text,
    unit: vscode_languageserver_1.CompletionItemKind.Unit,
    value: vscode_languageserver_1.CompletionItemKind.Value,
    variable: vscode_languageserver_1.CompletionItemKind.Variable
};
//# sourceMappingURL=typescript-service.js.map